@@ -1,11 +1,11 @@
 ;;; tempel.el --- Tempo templates/snippets with in-buffer field editing -*- lexical-binding: t -*-
 
-;; Copyright (C) 2022-2025 Free Software Foundation, Inc.
+;; Copyright (C) 2022-2026 Free Software Foundation, Inc.
 
 ;; Author: Daniel Mendler <mail@daniel-mendler.de>
 ;; Maintainer: Daniel Mendler <mail@daniel-mendler.de>
 ;; Created: 2022
-;; Version: 1.9
+;; Version: 1.10
 ;; Package-Requires: ((emacs "29.1") (compat "30"))
 ;; URL: https://github.com/minad/tempel
 ;; Keywords: abbrev, languages, tools, text
@@ -164,9 +164,9 @@ may be named with `tempel--name' or carry an evaluatable Lisp expression
 (defun tempel--print-template (template)
   "Print TEMPLATE."
   (cl-loop
-   for elt in template until (keywordp elt) concat
+   for elt in template until (keywordp elt) collect
    (pcase elt
-     ('nil nil)
+     ('nil "")
      ((pred stringp) elt)
      (`(s ,name) (symbol-name name))
      (`(,(or 'p 'P) ,_ ,name . ,noinsert)
@@ -174,7 +174,8 @@ may be named with `tempel--name' or carry an evaluatable Lisp expression
      ('> " ")
      ('n> "\n ")
      ((or 'n '& '% 'o) "\n")
-     (_ #("_" 0 1 (face shadow))))))
+     (_ #("_" 0 1 (face shadow))))
+   into result finally return (apply #'concat result)))
 
 (defun tempel--template-plist (template)
   "Get property list from TEMPLATE list."
@@ -182,7 +183,7 @@ may be named with `tempel--name' or carry an evaluatable Lisp expression
 
 (defun tempel--insert-doc (template)
   "Insert documentation of TEMPLATE."
-  (when-let ((doc (plist-get (tempel--template-plist template) :doc)))
+  (when-let* ((doc (plist-get (tempel--template-plist template) :doc)))
     (unless (eq ?\n (char-before)) (insert "\n"))
     (insert (propertize "\n" 'face '(:overline t :height 0.1 :extend t)))
     (insert doc)))
@@ -190,14 +191,14 @@ may be named with `tempel--name' or carry an evaluatable Lisp expression
 (defun tempel--annotate (templates width sep name)
   "Annotate template NAME given the list of TEMPLATES.
 WIDTH and SEP configure the formatting."
-  (when-let ((name (intern-soft name))
-             (elts (cdr (assoc name templates))))
+  (when-let* ((name (intern-soft name))
+              (template (alist-get name templates)))
     (let ((ann (truncate-string-to-width
                 (string-trim
                  (replace-regexp-in-string
                   "[ \t\n\r]+" " "
-                  (or (plist-get (tempel--template-plist elts) :ann)
-                      (tempel--print-template elts))))
+                  (or (plist-get (tempel--template-plist template) :ann)
+                      (tempel--print-template template))))
                 width)))
       (add-face-text-property 0 (length ann) 'completions-annotations t ann)
       (concat sep ann))))
@@ -206,13 +207,13 @@ WIDTH and SEP configure the formatting."
   "Create info buffer for template NAME.
 FUN inserts the info into the buffer.
 TEMPLATES is the list of templates."
-  (when-let ((name (intern-soft name))
-             (elts (cdr (assoc name templates))))
+  (when-let* ((name (intern-soft name))
+              (template (alist-get name templates)))
     (with-current-buffer (get-buffer-create " *tempel-info*")
       (setq buffer-read-only t)
       (let ((inhibit-read-only t))
         (erase-buffer)
-        (funcall fun elts)))))
+        (funcall fun template)))))
 
 (defun tempel--delete-word (word)
   "Delete WORD before point."
@@ -224,15 +225,16 @@ TEMPLATES is the list of templates."
   "Exit function for completion for template NAME and STATUS.
 TEMPLATES is the list of templates.
 REGION are the current region bounds."
-  (when-let (((not (eq status 'exact)))
-             (sym (intern-soft name))
-             (template (alist-get sym templates)))
+  (when-let* (((not (eq status 'exact)))
+              (sym (intern-soft name))
+              (template (alist-get sym templates)))
     (tempel--delete-word name)
     (tempel--insert template region)))
 
 (defun tempel--range-modified (ov &rest _)
   "Range overlay OV modified."
-  (when (and (not tempel--inhibit-hooks) (= (overlay-start ov) (overlay-end ov)))
+  (when (and (not tempel--inhibit-hooks)
+             (= (overlay-start ov) (overlay-end ov)))
     (let ((inhibit-modification-hooks nil)
           (tempel--inhibit-hooks t))
       (tempel--disable (overlay-get ov 'tempel--range)))))
@@ -255,15 +257,15 @@ BEG and END are the boundaries of the modification."
         (let ((st (overlay-get ov 'tempel--field)))
           (unless undo-in-progress
             (move-overlay ov (overlay-start ov) (max end (overlay-end ov))))
-          (when-let ((name (overlay-get ov 'tempel--name)))
+          (when-let* ((name (overlay-get ov 'tempel--name)))
             (setf (alist-get name (cdr st))
                   (buffer-substring-no-properties
                    (overlay-start ov) (overlay-end ov))))
           (unless undo-in-progress
-            (tempel--synchronize-fields st ov)))))
+            (tempel--sync-fields st ov)))))
       (tempel--update-mark ov))))
 
-(defun tempel--synchronize-fields (st current)
+(defun tempel--sync-fields (st current)
   "Synchronize fields of ST, except CURRENT overlay."
   (let ((range (caar st)))
     (dolist (ov (cdar st))
@@ -276,20 +278,20 @@ BEG and END are the boundaries of the modification."
                         (and (setq x (overlay-get ov 'tempel--name))
                              (alist-get x (cdr st)))))
             (when x
-              (tempel--synchronize-replace (overlay-start ov)
+              (tempel--sync-replace (overlay-start ov)
                                            (overlay-end ov) ov x)))))
       ;; Move range overlay
       (move-overlay range (overlay-start range)
                     (max (overlay-end range) (overlay-end ov))))))
 
-(defun tempel--synchronize-replace (beg end ov str)
+(defun tempel--sync-replace (beg end ov str)
   "Replace region between BEG and END with STR.
 If OV is alive, move it."
   (let ((old (buffer-substring-no-properties beg end)))
     (setq ov (and ov (overlay-buffer ov) ov))
     (unless (equal str old)
       (unless (eq buffer-undo-list t)
-        (push (list 'apply #'tempel--synchronize-replace
+        (push (list 'apply #'tempel--sync-replace
                     beg (+ beg (length str)) ov old)
               buffer-undo-list))
       (let ((buffer-undo-list t))
@@ -311,12 +313,14 @@ If OV is alive, move it."
                  (and (= (overlay-start ov) (overlay-end ov))
                       tempel-mark))))
 
-(defun tempel--field (st &optional name init)
-  "Add template field to ST.
+(defun tempel--field (&optional name init default)
+  "Add template field.
 NAME is the optional field name.
 INIT is the optional initial input.
+DEFAULT specifies if initial input is a default value.
 Return the added field."
-  (let ((ov (make-overlay (point) (point)))
+  (let ((st (car tempel--active))
+        (ov (make-overlay (point) (point)))
         (hooks (list #'tempel--field-modified)))
     (push ov (car st))
     (when name
@@ -332,34 +336,82 @@ Return the added field."
     (overlay-put ov 'insert-in-front-hooks hooks)
     (overlay-put ov 'insert-behind-hooks hooks)
     (overlay-put ov 'face 'tempel-field)
-    (when (and init (get-text-property 0 'tempel--default init))
+    (when (and init default)
       (overlay-put ov 'face 'tempel-default)
       (overlay-put ov 'tempel--default
                    (if (string-suffix-p ": " init) 'end 'start)))
-    (tempel--synchronize-fields st ov)
+    (tempel--sync-fields st ov)
     ov))
 
-(defun tempel--form (st form)
-  "Add new template field evaluating FORM to ST.
+(defun tempel--form (form initial)
+  "Add new template field evaluating FORM with INITIAL text.
 Return the added field."
-  (let ((beg (point)))
-    (condition-case nil
-        (insert (or (eval form (cdr st)) ""))
-      ;; Ignore errors since some variables may not be defined yet.
-      (void-variable nil))
-    (let ((ov (make-overlay beg (point) nil t)))
-      (overlay-put ov 'face 'tempel-form)
-      (overlay-put ov 'tempel--form form)
-      (push ov (car st))
-      ov)))
+  (insert initial)
+  (let ((ov (make-overlay (- (point) (length initial)) (point) nil t)))
+    (overlay-put ov 'face 'tempel-form)
+    (overlay-put ov 'tempel--form form)
+    (push ov (caar tempel--active))
+    ov))
 
 (defmacro tempel--protect (&rest body)
   "Protect BODY, catch errors."
   `(with-demoted-errors "Tempel Error: %S"
      ,@body))
 
-(defun tempel--element (st region elt)
-  "Add template ELT to ST given the REGION."
+(defun tempel--element (region elt)
+  "Add template ELT given the REGION.
+A template can consist of elements of several types:
+
+- string: The string is inserted in the buffer.
+- nil: It is ignored.
+- `p': An empty and unnamed placeholder field is inserted.
+- `r': Inserts the currently active region.  If no region is active, a
+  placeholder field is inserted.  If `tempel-done-on-region' is non-nil,
+  the template is finished when you jump to the field like `q'.
+- `r>': Like `r', but it also indents the region.
+- `n': Inserts a newline.
+- `n>': Inserts a newline and indents line.
+- `>': The line is indented using `indent-according-to-mode'.  Note
+  that you often should place this item after the text you want on the
+  line.
+- `&': If there is only whitespace between the line start and point,
+  nothing happens.  Otherwise a newline is inserted.
+- `%': If there is only whitespace between point and end of line,
+  nothing happens.  Otherwise a newline is inserted.
+- `o': Like `%' but leaves the point before the newline.
+- (s NAME): Inserts a named field.
+- (p PROMPT <NAME> <NOINSERT>): Insert an optionally named field with a
+  prompt.  The PROMPT is displayed directly in the buffer as default
+  value.  The field value is bound to NAME and updated dynamically.  If
+  NOINSERT is non-nil, no field is inserted and the minibuffer is used
+  for prompting.  For clarity, the symbol `noinsert' should be used as
+  argument.
+- (r PROMPT <NAME> <NOINSERT>): Like (p ..), but if there is a current
+  region, it is placed here.
+- (r> PROMPT <NAME> <NOINSERT>): Like (r ..), but is also indents the
+  region.
+- (l ELEMENTS..): Insert multiple elements.
+- Anything else is passed to each function in `tempel-user-elements'
+  until one of the functions returns non-nil, and the result is
+  inserted.  If all of them return nil, the form is evaluated.  The
+  result can either be a string or any other element.  If the return
+  value is a string it is dynamically updated on modification of other
+  fields.  Other return values are treated as elements and inserted
+  according to the rules.  The element (l ..) is useful to return
+  multiple elements.
+
+Tempel extends the Tempo syntax with the following elements:
+
+- (p FORM <NAME> <NOINSERT>): Like (p ..) described above, but FORM is
+  evaluated.  FORM can for example call `completing-read' to select
+  among various elements.
+- (FORM ..): If a Lisp form evaluates to a string, it is inserted as
+  overlay and the overlay is updated on modifications of other fields.
+- `q': Like `p', but the template is finished if the user jumps to the
+  field.  Similarly `r' finishes the template if `tempel-done-on-region'
+  is non-nil.
+
+Use caution with templates which execute arbitrary code!"
   (pcase elt
     ('nil)
     ('n (insert "\n"))
@@ -373,51 +425,60 @@ Return the added field."
           (insert "\n")))
     ('o (unless (or (eolp) (save-excursion (re-search-forward "\\=\\s-*$" nil t)))
           (open-line 1)))
-    (`(s ,name) (tempel--field st name))
-    (`(l . ,lst) (dolist (e lst) (tempel--element st region e)))
-    ((or 'p `(,(or 'p 'P) . ,rest)) (apply #'tempel--placeholder st rest))
+    (`(s ,name) (tempel--field name))
+    (`(l . ,lst) (dolist (e lst) (tempel--element region e)))
+    ((or 'p `(,(or 'p 'P) . ,rest)) (apply #'tempel--placeholder rest))
     ((or 'r 'r> `(,(or 'r 'r>) . ,rest))
      (if (not region)
-         (when-let ((ov (apply #'tempel--placeholder st rest))
-                    ((not rest))
-                    (tempel-done-on-region))
+         (when-let* ((ov (apply #'tempel--placeholder rest))
+                     ((not rest))
+                     (tempel-done-on-region))
            (overlay-put ov 'tempel--enter #'tempel--done))
        (goto-char (cdr region))
        (when (eq (or (car-safe elt) elt) 'r>)
          (indent-region (car region) (cdr region) nil))))
     ;; TEMPEL EXTENSION: Quit template immediately
-    ('q (overlay-put (tempel--field st) 'tempel--enter #'tempel--done))
-    (_ (if-let ((ret (run-hook-wrapped 'tempel-user-elements
-                                       (lambda (hook elt fields)
-                                         (condition-case nil
-                                             (funcall hook elt)
-                                           (wrong-number-of-arguments
-                                            (funcall hook elt fields))))
-                                       elt (cdr st))))
-           (tempel--element st region ret)
-         ;; TEMPEL EXTENSION: Evaluate forms
-         (tempel--form st elt)))))
+    ('q (overlay-put (tempel--field) 'tempel--enter #'tempel--done))
+    (_ (let* ((uel (tempel--user-element elt))
+              (val (unless uel
+                      ;; Ignore errors since variables may not be defined yet.
+                      (condition-case nil
+                          (eval elt (cdar tempel--active))
+                        (void-variable "")))))
+         (if (or uel (not (stringp val)))
+             (tempel--element region (or uel val))
+           ;; TEMPEL EXTENSION: Evaluate forms
+           (tempel--form elt val))))))
 
-(defun tempel--placeholder (st &optional prompt name noinsert)
-  "Handle placeholder element and add field with NAME to ST.
+(defun tempel--user-element (elt)
+  "Evaluate user element ELT."
+  (run-hook-wrapped 'tempel-user-elements
+                    (lambda (hook elt fields)
+                      (condition-case nil
+                          (funcall hook elt)
+                        (wrong-number-of-arguments
+                         (funcall hook elt fields))))
+                    elt (cdar tempel--active)))
+
+(defun tempel--placeholder (&optional prompt name noinsert)
+  "Handle placeholder element and add field with NAME.
 If NOINSERT is non-nil do not insert a field, only bind the value to NAME.
 PROMPT is the optional prompt/default value.
 If a field was added, return it."
-  (setq prompt
-        (cond
-         ((and (stringp prompt) noinsert) (read-string prompt))
-         ((stringp prompt) (propertize prompt 'tempel--default t))
-         ;; TEMPEL EXTENSION: Evaluate prompt
-         (t (eval prompt (cdr st)))))
-  (if noinsert
-      (progn (setf (alist-get name (cdr st)) prompt) nil)
-    (tempel--field st name prompt)))
+  (let ((init
+         (cond
+          ((and (stringp prompt) noinsert) (read-string prompt))
+          ((stringp prompt) prompt)
+          ;; TEMPEL EXTENSION: Evaluate prompt
+          (t (eval prompt (cdar tempel--active))))))
+    (if noinsert
+        (progn (setf (alist-get name (cdar tempel--active)) init) nil)
+      (tempel--field name init (stringp prompt)))))
 
 (defun tempel--insert (template region)
   "Insert TEMPLATE given the current REGION."
   (let ((plist (tempel--template-plist template)))
     (eval (plist-get plist :pre) 'lexical)
-    ;; TODO do we want to have the ability to reactivate snippets?
     (unless (eq buffer-undo-list t)
       (push '(apply tempel--disable) buffer-undo-list))
     (setf (alist-get 'tempel--active minor-mode-overriding-map-alist) tempel-map)
@@ -425,21 +486,22 @@ If a field was added, return it."
       ;; Split existing overlays, do not expand within existing field.
       (dolist (st tempel--active)
         (dolist (ov (cdar st))
-          (when (and (<= (overlay-start ov) (point)) (>= (overlay-end ov) (point)))
+          (when (and (<= (overlay-start ov) (point))
+                     (>= (overlay-end ov) (point)))
             (setf (overlay-end ov) (point)))))
       ;; Activate template
       (let ((st (cons nil nil))
-            (ov (point))
+            (beg (point))
             (tempel--inhibit-hooks t))
+        (push st tempel--active)
         (cl-loop for x in template until (keywordp x)
-                 do (tempel--element st region x))
-        (setq ov (make-overlay ov (point) nil t))
-        (push ov (car st))
-        (overlay-put ov 'modification-hooks (list #'tempel--range-modified))
-        (overlay-put ov 'tempel--range st)
-        (overlay-put ov 'tempel--post (plist-get plist :post))
-        ;;(overlay-put ov 'face 'region) ;; For debugging
-        (push st tempel--active)))
+                 do (tempel--element region x))
+        (let ((ov (make-overlay beg (point) nil t)))
+          (overlay-put ov 'modification-hooks (list #'tempel--range-modified))
+          (overlay-put ov 'tempel--range st)
+          (overlay-put ov 'tempel--post (plist-get plist :post))
+          ;; (overlay-put ov 'face 'region) ;; Enable for debugging
+          (push ov (car st)))))
     (cond
      ((cl-loop for ov in (caar tempel--active)
                never (overlay-get ov 'tempel--field))
@@ -448,14 +510,14 @@ If a field was added, return it."
      ((cl-loop for ov in (caar tempel--active)
                never (and (overlay-get ov 'tempel--field)
                           (eq (point) (overlay-start ov))))
-      (tempel-next 1))))) ;; Jump to first field
+      (tempel-next))))) ;; Jump to first field
 
 (defun tempel--save ()
   "Prompt to save modified files in `tempel-path'."
   (cl-loop
    with all = nil
    for (file . _ts) in (car tempel--path-templates) do
-   (when-let ((buf (get-file-buffer file)))
+   (when-let* ((buf (get-file-buffer file)))
      (with-current-buffer buf
        (when (and (buffer-modified-p)
                   (pcase (or all (read-answer
@@ -480,27 +542,31 @@ If a field was added, return it."
 
 (defun tempel--file-prepare (data)
   "Reorganize the template DATA from file.
-DATA must be a list (modes plist templates modes plist templates...).
+DATA must be a list (MODES PLIST TEMPLATES MODES PLIST TEMPLATES...).
 See the README, which provides an example for the file format of the
-template.eld file.  The return value is a list of (modes plist . templates)."
+template.eld file.  The return value is a list of elements of the
+form (MODE COND . TEMPLATES)."
   (let (result)
     (while data
       (let (modes plist templates)
         (while (and (car data) (symbolp (car data)) (not (keywordp (car data))))
           (push (pop data) modes))
         (while (keywordp (car data))
-          (push (pop data) plist)
-          (push (pop data) plist))
+          (setq plist `(,(car data) ,(cadr data) ,@plist)
+                data (cddr data)))
         (while (consp (car data))
           (push (pop data) templates))
-        (push `(,(nreverse modes) ,(nreverse plist) . ,(nreverse templates)) result)))
+        (setq plist (or (not (plist-member plist :when)) (plist-get plist :when))
+              templates (nreverse templates))
+        (dolist (mode modes)
+          (push `(,mode ,plist ,@templates) result))))
     result))
 
 (defun tempel-path-templates ()
   "Return templates defined in `tempel-path'.
-Additionally, save any files in `tempel-template-sources' that have been
-modified since the last time this function was called.
-This is meant to be a source in `tempel-template-sources'."
+Additionally, save any files in `tempel-path' that have been modified
+since the last time this function was called.  This function can be used
+as source in `tempel-template-sources'."
   (when (or (not tempel--path-templates) tempel-auto-reload)
     (let* ((files
             (cl-loop for path in (ensure-list tempel-path)
@@ -512,45 +578,43 @@ This is meant to be a source in `tempel-template-sources'."
                                     (file-attributes (file-chase-links f)))
                                    'integer)))))
       (unless (equal (car tempel--path-templates) timestamps)
-        (setq tempel--path-templates (cons timestamps
-                                           (mapcan #'tempel--file-read files))))))
+        (setq tempel--path-templates
+              (cons timestamps (mapcan #'tempel--file-read files))))))
   (tempel--filter-templates (cdr tempel--path-templates)))
 
 (defun tempel--filter-templates (templates)
   "Filter templates from TEMPLATES relevant to the current buffer.
-TEMPLATES must be a list in the form (modes plist . templates)."
-  (cl-loop for (modes plist . mode-templates) in templates
-           if (tempel--condition-p modes plist)
+TEMPLATES must be a list of elements of the form (MODE COND . TEMPLATES)."
+  (cl-loop for (mode cond . mode-templates) in templates
+           if (tempel--condition-p mode cond)
            append mode-templates))
 
-(defun tempel--condition-p (modes plist)
-  "Return non-nil if one of MODES matches and the PLIST condition is satisfied."
+(defun tempel--condition-p (mode cond)
+  "Return non-nil if MODE matches and COND is satisfied."
   (and
-   (cl-loop
-    for m in modes thereis
-    (or (eq m #'fundamental-mode)
-        (derived-mode-p m)
-        (when-let ((remap (alist-get m major-mode-remap-alist)))
-          (derived-mode-p remap))))
+   (or (eq mode #'fundamental-mode)
+       (derived-mode-p mode)
+       (when-let* ((remap (alist-get mode major-mode-remap-alist)))
+         (derived-mode-p remap)))
    (or tempel--ignore-condition
-       (not (plist-member plist :when))
+       (eq cond t)
        (save-excursion
          (save-restriction
            (save-match-data
-             (eval (plist-get plist :when) 'lexical)))))))
+             (eval cond 'lexical)))))))
 
 (defun tempel--templates ()
   "Return templates for current mode."
-  (let (result)
+  (let (list)
     (run-hook-wrapped
      'tempel-template-sources
      (lambda (fun)
        (cond
-        ((functionp fun) (cl-callf append result (funcall fun)))
-        ((boundp fun) (cl-callf append result (symbol-value fun)))
+        ((functionp fun) (push (funcall fun) list))
+        ((boundp fun) (push (symbol-value fun) list))
         (t (error "Template source is not a function or a variable: %S" fun)))
        nil))
-    result))
+    (apply #'append list)))
 
 (defun tempel--region ()
   "Return region bounds."
@@ -565,8 +629,10 @@ TEMPLATES must be a list in the form (modes plist . templates)."
     (dolist (st tempel--active next)
       (dolist (ov (car st))
         (unless (overlay-get ov 'tempel--form)
-          (setq stop (if (or (< dir 0) (eq 'start (overlay-get ov 'tempel--default)))
-                         (overlay-start ov) (overlay-end ov)))
+          (setq stop (if (or (< dir 0)
+                             (eq 'start (overlay-get ov 'tempel--default)))
+                         (overlay-start ov)
+                       (overlay-end ov)))
           (cond
            ((and (> dir 0) (> stop pt))
             (setq next (min (or next (point-max)) stop)))
@@ -581,7 +647,7 @@ TEMPLATES must be a list in the form (modes plist . templates)."
   "Move to beginning of the template."
   (declare (completion tempel--active-p))
   (interactive)
-  (when-let ((pos (tempel--beginning)))
+  (when-let* ((pos (tempel--beginning)))
     (cond
      ((/= pos (point)) (goto-char pos))
      (tempel-done-on-next (tempel-done t)))))
@@ -590,7 +656,7 @@ TEMPLATES must be a list in the form (modes plist . templates)."
   "Move to end of the template."
   (declare (completion tempel--active-p))
   (interactive)
-  (when-let ((pos (tempel--end)))
+  (when-let* ((pos (tempel--end)))
     (cond
      ((/= pos (point)) (goto-char pos))
      (tempel-done-on-next (tempel-done t)))))
@@ -608,30 +674,33 @@ TEMPLATES must be a list in the form (modes plist . templates)."
   "Kill the field contents."
   (declare (completion tempel--active-p))
   (interactive)
-  (if-let ((ov (tempel--find-overlay 'tempel--field)))
+  (if-let* ((ov (tempel--find-overlay 'tempel--field)))
       (kill-region (overlay-start ov) (overlay-end ov))
     (kill-sentence nil)))
 
-(defun tempel-next (arg)
-  "Move ARG fields forward and quit at the end."
+(defun tempel-next (&optional n)
+  "Move N fields forward.
+Quit at the end if `tempel-done-on-next' is non-nil."
   (declare (completion tempel--active-p))
   (interactive "p")
-  (cl-loop for i below (abs arg) do
-           (if-let ((next (tempel--find arg)))
+  (setq n (or n 1))
+  (cl-loop for i below (abs n) do
+           (if-let* ((next (tempel--find n)))
                (goto-char next)
              (when tempel-done-on-next
                (tempel-done t))
              (cl-return)))
   ;; Run the enter action of the field.
-  (when-let ((ov (tempel--find-overlay 'tempel--field))
-             (fun (overlay-get ov 'tempel--enter)))
+  (when-let* ((ov (tempel--find-overlay 'tempel--field))
+              (fun (overlay-get ov 'tempel--enter)))
     (funcall fun ov)))
 
-(defun tempel-previous (arg)
-  "Move ARG fields backward and quit at the beginning."
+(defun tempel-previous (&optional n)
+  "Move N fields backward.
+Quit at the beginning if `tempel-done-on-next' is non-nil."
   (declare (completion tempel--active-p))
   (interactive "p")
-  (tempel-next (- arg)))
+  (tempel-next (- (or n 1))))
 
 (defun tempel--beginning ()
   "Return beginning of template markers."
@@ -696,9 +765,17 @@ If prefix argument ALL is given, abort all templates."
     (with-current-buffer buf
       (tempel--disable st))))
 
-(defun tempel--prefix-bounds ()
-  "Return prefix bounds."
-  (bounds-of-thing-at-point 'symbol))
+(defun tempel--prefix-bounds (templates)
+  "Return prefix bounds given TEMPLATES list."
+  (let ((beg (save-excursion (skip-chars-backward "^[:space:]") (point)))
+        (end (point)))
+    (if (and (/= beg end)
+             ;; Check if prefix matches a template name.
+             (try-completion (buffer-substring-no-properties beg end)
+                             templates))
+        (cons beg end)
+      ;; Fallback to `bounds-of-thing-at-point'.
+      (bounds-of-thing-at-point 'symbol))))
 
 ;;;###autoload
 (defun tempel-expand (&optional interactive)
@@ -713,12 +790,12 @@ command."
   (interactive (list t))
   (when interactive
     (tempel--save))
-  (if-let ((templates (tempel--templates))
-           (bounds (bounds-of-thing-at-point 'symbol))
-           (name (buffer-substring-no-properties
-                  (car bounds) (cdr bounds)))
-           (sym (intern-soft name))
-           (template (assq sym templates)))
+  (if-let* ((templates (tempel--templates))
+            (bounds (tempel--prefix-bounds templates))
+            (name (buffer-substring-no-properties
+                   (car bounds) (cdr bounds)))
+            (sym (intern-soft name))
+            (template (assq sym templates)))
       (if interactive
           (tempel--exit templates nil name 'finished)
         (setq templates (list template))
@@ -745,9 +822,10 @@ Capf, otherwise like an interactive completion command."
           (user-error "tempel-complete: No matching templates")))
     ;; Use the marked region for template insertion if triggered manually.
     (let ((region (and (eq this-command #'tempel-complete) (tempel--region))))
-      (when-let ((templates (tempel--templates))
-                 (bounds (or (and (not region) (bounds-of-thing-at-point 'symbol))
-                             (cons (point) (point)))))
+      (when-let* ((templates (tempel--templates))
+                  (bounds (or (and (not region)
+                                   (tempel--prefix-bounds templates))
+                              (cons (point) (point)))))
         (list (car bounds) (cdr bounds) templates
               :category 'tempel
               :exclusive 'no
@@ -755,22 +833,22 @@ Capf, otherwise like an interactive completion command."
               :exit-function (apply-partially #'tempel--exit templates region)
               :company-doc-buffer
               (apply-partially #'tempel--info-buffer templates
-                               (lambda (elts)
-                                 (insert (tempel--print-template elts))
-                                 (tempel--insert-doc elts)
+                               (lambda (template)
+                                 (insert (tempel--print-template template))
+                                 (tempel--insert-doc template)
                                  (current-buffer)))
               :company-location
               (apply-partially #'tempel--info-buffer templates
-                               (lambda (elts)
-                                 (pp (cl-loop for x in elts
+                               (lambda (template)
+                                 (pp (cl-loop for x in template
                                               until (keywordp x) collect x)
                                      (current-buffer))
-                                 (tempel--insert-doc elts)
+                                 (tempel--insert-doc template)
                                  (list (current-buffer))))
               :annotation-function
-              (and tempel-complete-annotation
-                   (apply-partially #'tempel--annotate
-                                    templates tempel-complete-annotation " ")))))))
+              (when tempel-complete-annotation
+                (apply-partially #'tempel--annotate
+                                 templates tempel-complete-annotation " ")))))))
 
 ;;;###autoload
 (defun tempel-insert (template-or-name)
@@ -786,6 +864,7 @@ If called interactively, select a template with `completing-read'."
                (intern-soft
                 (completing-read
                  "Template: "
+                 ;; TODO: Use `completion-table-with-metadata' via Compat 31
                  (lambda (str pred action)
                    (if (eq action 'metadata)
                        `(metadata
@@ -832,17 +911,16 @@ If called interactively, select a template with `completing-read'."
   (when tempel-abbrev-mode
     (let ((table (make-abbrev-table))
           (tempel--ignore-condition t))
-      (dolist (template (tempel--templates))
-        (let* ((name (symbol-name (car template)))
-               (hook (make-symbol name)))
+      (dolist (sym (delete-dups (mapcar #'car (tempel--templates))))
+        (let ((hook (make-symbol (symbol-name sym))))
           (fset hook (lambda ()
-                       (tempel--delete-word name)
-                       (tempel--insert (cdr template) nil)
+                       (tempel--delete-word (symbol-name sym))
+                       (tempel--insert (alist-get sym (tempel--templates)) nil)
                        t))
           (put hook 'no-self-insert t)
-          (define-abbrev table name 'Template hook
+          (define-abbrev table (symbol-name sym) 'Template hook
             :system t :enable-function
-            (lambda () (assq (car template) (tempel--templates))))))
+            (lambda () (assq sym (tempel--templates))))))
       (setq-local abbrev-minor-mode-table-alist
                   (cons `(tempel-abbrev-mode . ,table)
                         abbrev-minor-mode-table-alist)))))
